PROGRAM Protpars (infile, OUTPUT);
  (* version 2.6 *)

CONST maxsp = 50;    (* maximum number of species               *)
   maxsz = 100;      (* size of pointer array.  >= 2*maxsp - 1  *)
   maxsites = 200;   (* max. number amino acids per sequence    *)
   nmlngth = 10;     (* number of characters in species name    *)

TYPE steparray = ARRAY[1..maxsites] OF INTEGER;
   aas = (ala, arg, asn, asp, cys, gln, glu, gly, his, ileu, leu, lys,
          met, phe, pro, ser1, ser2, thr, trp, tyr, val, del, stop,
          asx, ser, unk);
   aaset = SET OF ala..stop;
   sitearray = ARRAY[0..3] OF aaset;
   pointer = ^node;           (* nodes will form binary tree *)
   node = RECORD      (* describes a tip species or an ancestor *)
      ancestor, leftdesc, rtdesc : pointer;
      index : INTEGER;
      tip : BOOLEAN;        (* present species are tips of tree *)
      seq : ARRAY [1..maxsites] OF aas; (* the sequence *)
      siteset : sitearray;  (* temporary storage for aa's *)
      xcoord, ycoord : INTEGER     (* used by printree *)
      END;
   pointarray = ARRAY [1..maxsz] OF pointer;

VAR root : pointer;
   infile : TEXT;
   spp, nonodes, chars, numtrees, outgrno : INTEGER;
     (* spp = number of species
        nonodes = number of nodes in tree
        chars = number of sites in actual sequences
        numtrees is used by usertree option part of maketree
        outgrno is used by reroot *)
   brief, usertree, weights, thresh, outgropt, didreroot : BOOLEAN;
   weight : steparray;
   treenode : pointarray;      (* pointers to all nodes in tree *)
   name : ARRAY[1..maxsp, 1..nmlngth] OF CHAR; (* names of species *)
   threshold : REAL;

   FUNCTION digit (ch : CHAR) : BOOLEAN;
   BEGIN (* digit *)
      digit := (ch >= '0') AND (ch <= '9');
   END; (* digit *)

   FUNCTION letter (ch : CHAR) : BOOLEAN;
   BEGIN (* letter -- works for both ASCII and EBCDIC codes *)
      letter := ((ch >= 'A') AND (ch <= 'I'))
                OR ((ch >= 'J') AND (ch <= 'R'))
                OR ((ch >= 'S') AND (ch <= 'Z'))
                OR ((ch >= 'a') AND (ch <= 'i'))
                OR ((ch >= 'j') AND (ch <= 'r'))
                OR ((ch >= 's') AND (ch <= 'z'));
   END; (* letter *)

   PROCEDURE uppercase (VAR ch : CHAR);
   BEGIN (* uppercase *)
      IF letter(ch) AND (ch >= 'a')
      THEN ch := CHR(ORD(ch) + ORD('A') - ORD('a'));
   END; (* uppercase *)

   PROCEDURE doinput;
      (* reads the input data *)

      PROCEDURE inputnumbers;
         (* input the numbers of species and of characters *)
      BEGIN (* inputnumbers *)
         WRITELN(OUTPUT);
         WRITE(OUTPUT, ' Protein sequence parsimony algorithm');
         WRITELN(OUTPUT, ' version 2.6');
         WRITELN(OUTPUT);
         READ(infile, spp, chars);
         WRITELN(OUTPUT, ' ',spp:2,' species, ',chars:3,'  sites');
         WRITELN(OUTPUT);
         nonodes := 2 * spp - 1;
      END; (* inputnumbers *)

      PROCEDURE inputoptions;
         (* input the information on the options *)
         VAR ch : CHAR;
            extranum, i : INTEGER;

         PROCEDURE inputhresh;
            (* input the threshold *)
         BEGIN (* inputhresh *)
            READLN(infile, threshold);
            thresh := TRUE;
         END; (* inputhresh *)

         PROCEDURE printhresh;
            (* print out the threshold value *)
         BEGIN (* printhresh *)
            WRITELN(OUTPUT, '    Threshold value = ', threshold:10:3);
            WRITELN(OUTPUT);
         END; (* printhresh *)

         PROCEDURE inputweights;
            (* input the character weights, 0-9 and A-Z for weights
               0 - 36 *)
            VAR ch : CHAR;
               i : INTEGER;
         BEGIN (* inputweights *)
            FOR i := 1 TO nmlngth-1 DO READ(infile, ch);
            FOR i := 1 TO chars DO BEGIN
               IF EOLN(infile) THEN READLN(infile);
               READ(infile, ch);
               weight[i] := 1;
               IF digit(ch)
               THEN weight[i] := ORD(ch) - ORD('0')
               ELSE IF letter(ch)
                  THEN BEGIN
                     uppercase (ch);
                     IF (ch >= 'A') AND (ch <= 'I')
                     THEN weight[i] := ORD(ch) - ORD('A') + 10
                     ELSE IF (ch >= 'J') AND (ch <= 'R')
                        THEN weight[i] := ORD(ch) - ORD('J') + 19
                        ELSE weight[i] := ORD(ch) - ORD('S') + 28;
                     END
                  ELSE WRITELN(OUTPUT, ' Bad weight character: ',ch);
               END;
            READLN(infile);
            weights := TRUE;
         END; (* inputweights *)

         PROCEDURE printweights;
            (* print out the weights of sites *)
            VAR i, j, k : INTEGER;
         BEGIN (* printweights *)
            WRITELN(OUTPUT,'    Sites are weighted as follows:');
            WRITE(OUTPUT, '        ');
            FOR i := 0 TO 9 DO WRITE(OUTPUT, i:3);
            WRITELN(OUTPUT);
            WRITELN(OUTPUT,'     *---------------------------------');
            FOR j := 0 TO (chars DIV 10) DO BEGIN
               WRITE(OUTPUT, 10*j:5, '!  ');
               FOR i := 0 TO 9 DO BEGIN
                  k := 10*j + i;
                  IF (k>0) AND (k<=chars)
                  THEN WRITE(OUTPUT, weight[k]:3)
                  ELSE WRITE(OUTPUT, '   ');
                  END;
               WRITELN(OUTPUT);
               END;
            WRITELN(OUTPUT);
         END; (* printweights *)

      BEGIN (* inputoptions *)
         brief := FALSE;
         usertree := FALSE;
         weights := FALSE;
         thresh := FALSE;
         outgropt := FALSE;
         extranum := 0;
         WHILE NOT EOLN(infile) DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF (ch = 'B') OR (ch = 'O') OR (ch = 'T') 
               OR (ch = 'U') OR (ch = 'W')
            THEN CASE ch OF
                  'B' : brief := TRUE;
                  'O' : BEGIN
                        outgropt := TRUE;
                        extranum := extranum + 1;
                        END;
                  'T' : extranum := extranum + 1;
                  'U' : usertree := TRUE;
                  'W' : extranum := extranum + 1
                  END
            ELSE IF ch <> ' '
               THEN WRITELN(OUTPUT, ' Bad Option Character: ', ch);
            END;
         READLN(infile);
         FOR i := 1 TO extranum DO BEGIN
            READ(infile, ch);
            uppercase (ch);
            IF ch = 'O'
            THEN READLN(infile, outgrno);
            IF ch = 'W'
            THEN inputweights;
            IF ch = 'T'
            THEN inputhresh;
            END;
         IF thresh
         THEN printhresh
         ELSE threshold := 35 * spp;
         IF weights
         THEN printweights
         ELSE FOR i := 1 TO chars DO weight[i] := 1;
      END; (* inputoptions *)

      PROCEDURE inputdata;
         (* input the names and sequences for each species *)
         VAR i, j, k : INTEGER;
            aa : aas; (* temporary amino acid for input *)
            charstate : CHAR;  (* character temp'y storage for input *)
      BEGIN (* inputdata *)
         WRITELN(OUTPUT, ' Sequences:');
         WRITELN(OUTPUT);
         FOR i := 1 TO nonodes DO BEGIN
            NEW (treenode[i]);
            treenode[i]^.leftdesc := NIL;
            treenode[i]^.rtdesc := NIL;
            treenode[i]^.ancestor := NIL;
            treenode[i]^.index := i;
            IF i <= spp
            THEN BEGIN
               treenode[i]^.tip := TRUE;
               FOR j := 1 TO nmlngth DO READ(infile, name[i,j]);
               WRITE(OUTPUT, ' ');
               FOR j := 1 TO nmlngth DO WRITE(OUTPUT, name[i,j]);
               WRITE(OUTPUT, '   ');
               FOR j := 1 TO chars DO BEGIN
                  IF EOLN(infile) THEN READLN(infile);
                  READ(infile, charstate);
                  uppercase (charstate);
                  IF (NOT letter (charstate)) AND
                     (NOT ((charstate = '.') OR (charstate = '?')
                      OR (charstate = '-')))
                     OR (charstate = 'J') OR (charstate = 'O')
                     OR (charstate = 'U') OR (charstate = 'Z')
                  THEN WRITELN(OUTPUT, ' Warning -- ',
                        'bad amino acid: ',charstate);
                  IF (j <> 1) AND ((j-1) MOD 50 = 0)
                  THEN BEGIN
                     WRITELN(OUTPUT);
                     FOR k := 1 TO nmlngth DO
                        WRITE(OUTPUT, ' ');
                     WRITE(OUTPUT, '    ');
                     END;
                  WRITE(OUTPUT, charstate);
                  IF (j MOD 5) = 0 THEN WRITE(OUTPUT, ' ');
                  CASE charstate OF
                     'A' : aa := ala;
                     'B' : aa := asx;
                     'C' : aa := cys;
                     'D' : aa := asp;
                     'E' : aa := glu;
                     'F' : aa := phe;
                     'G' : aa := gly;
                     'H' : aa := his;
                     'I' : aa := ileu;
                     'K' : aa := lys;
                     'L' : aa := leu;
                     'M' : aa := met;
                     'N' : aa := asn;
                     'P' : aa := pro;
                     'Q' : aa := gln;
                     'R' : aa := arg;
                     'S' : aa := ser;
                     'T' : aa := thr;
                     'V' : aa := val;
                     'W' : aa := trp;
                     'X' : aa := unk;
                     'Y' : aa := tyr;
                     '.' : aa := stop;
                     '?' : aa := unk;
                     '-' : aa := del
                     END;
                  treenode[i]^.seq[j] := aa;
                  END;
               READLN(infile);
               WRITELN(OUTPUT);
               END
            ELSE treenode[i]^.tip := FALSE;
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* inputdata *)

   BEGIN (* doinput *)
      inputnumbers;
      inputoptions;
      inputdata;
   END; (* doinput *)

   PROCEDURE maketree;
      (* constructs a binary tree from the pointers in treenode.
         adds each node at location which yields highest "likelihood"
         then rearranges the tree for greatest "likelihood" *)
      VAR i : INTEGER;
         like, bestyet : REAL;
         there, item : pointer;
         numsteps, besteps : steparray;
         translate : ARRAY[ala..unk] OF sitearray;

      PROCEDURE setup;
         (* set up set table to get aasets from aas *)
         VAR a, b : aas;
            s, fullset : aaset;
      BEGIN (* setup *)
         FOR a := ala TO stop DO
            translate [a][0] := [a];
         translate [ala][1] := [asp, glu, gly, ser1, pro, thr, val];
         translate [arg][1] := [cys, gln, gly, his, ileu, leu, lys,
                                met, pro, ser2, thr, trp, stop];
         translate [asn][1] := [asp, his, ileu, ser2, thr, tyr];
         translate [asp][1] := [ala, asn, glu, gly, his, tyr, val];
         translate [cys][1] := [arg, gly, phe, ser1, ser2, trp, tyr,
                                stop];
         translate [gln][1] := [arg, glu, his, leu, lys, pro, stop];
         translate [glu][1] := [ala, asp, gln, gly, lys, val, stop];
         translate [gly][1] := [ala, arg, asp, cys, glu, ser2,
                                trp, val, stop];
         translate [his][1] := [arg, asn, asp, gln, leu, pro, tyr];
         translate [ileu][1] := [arg, asn, leu, lys, met, phe, ser2,
                                 thr, val];
         translate [leu][1] := [arg, gln, his, ileu, met, phe,
                                pro, ser1, trp, val, stop];
         translate [lys][1] := [arg, asn, gln, glu, ileu, met, thr,
                                stop];
         translate [met][1] := [arg, ileu, leu, lys, val, thr];
         translate [phe][1] := [cys, ileu, leu, ser1, tyr, val];
         translate [pro][1] := [ala, arg, gln, his, leu, ser1, thr];
         translate [ser1][1] := [ala, cys, leu, phe, pro, thr, trp,
                                 tyr, stop];
         translate [ser2][1] := [arg, asn, cys, gly, ileu, thr];
         translate [thr][1] := [ala, arg, asn, ileu, lys, met,
                                pro, ser1, ser2];
         translate [trp][1] := [arg, cys, gly, leu, ser1, stop];
         translate [tyr][1] := [asn, asp, cys, his, phe, ser1, stop];
         translate [val][1] := [ala, asp, glu, gly, ileu,
                                leu, met, phe];
         translate [stop][1] := [arg, cys, gln, glu, gly, leu, lys,
                                 ser1, trp, tyr];
         translate [del][1] := [];
         fullset := [ala..stop] - [del];
         translate [unk][0] := fullset;
         translate [asx][0] := [asn, asp];
         translate [ser][0] := [ser1, ser2];
         translate [unk][1] := [];
         translate [asx][1] := translate [asn][1]
                                 + translate [asp][1] - [asn, asp];
         translate [ser][1] := translate [ser1][1]
                                + translate [ser2][1];
         FOR a := ala TO unk DO BEGIN
            s := [];
            FOR b := ala TO stop DO
               IF b IN translate [a][1]
               THEN s := s + translate [b][1];
            translate [a][2] := (s - translate [a][1])
                                    - translate [a][0];
            translate [a][3] := ((fullset - translate [a][1])
                                - translate [a][2]) - translate [a][0];
            END;
      END; (* setup *)

      PROCEDURE add (VAR below, newtip, newfork : pointer);
         (* inserts the nodes newfork and its left descendant, newtip,
            to the tree.  below becomes newfork's right descendant *)
      BEGIN (* add *)
         newfork^.ancestor := below^.ancestor;
         newfork^.rtdesc := below;
         newfork^.leftdesc := newtip;
         below^.ancestor := newfork;
         newtip^.ancestor := newfork;
         IF newfork^.ancestor <> NIL
         THEN IF newfork^.ancestor^.leftdesc = below
            THEN newfork^.ancestor^.leftdesc := newfork
            ELSE newfork^.ancestor^.rtdesc := newfork;
      END; (* add *)

      PROCEDURE remove (VAR item, fork : pointer);
         (* removes nodes item and its ancestor, fork, from the tree.
            the new descendant of fork's ancestor is made to be
            fork's second descendant (other than item).  Also
            returns pointers to the deleted nodes, item and fork *)
         VAR other, below : pointer;
      BEGIN (* remove *)
         IF item^.ancestor = NIL
         THEN fork := NIL
         ELSE BEGIN
            fork := item^.ancestor;
            IF item = fork^.leftdesc
            THEN other := fork^.rtdesc
            ELSE other := fork^.leftdesc;
            below := fork^.ancestor;
            other^.ancestor := below;
            IF below <> NIL
            THEN BEGIN
               IF below^.leftdesc = fork
               THEN below^.leftdesc := other
               ELSE below^.rtdesc := other;
               fork^.ancestor := NIL;
               END;
            fork^.leftdesc := NIL;
            fork^.rtdesc := NIL;
            item^.ancestor := NIL;
            END;
      END; (* remove *)

      PROCEDURE evaluate (r: pointer);
         (* determines the number of steps needed for a tree. this is
            the minimum number of steps needed to evolve sequences on
            this tree *)
         VAR i, m, steps : INTEGER;
            sum : REAL;

         PROCEDURE postorder (p : pointer);
            (* traverses a binary tree, calling procedure fillin at a
               node's descendants before calling fillin at the node *)

            PROCEDURE fillin (q: pointer);
               (* sets up for each node in the tree the aa set for site m
                  at that point and counts the changes *)
               VAR counted : BOOLEAN;
                  a : aas;
                  s : aaset;
                  l, r : pointer;
                  i, j, k, n : INTEGER;
            BEGIN (* fillin *)
               counted := FALSE;
               l := q^.leftdesc;
               r := q^.rtdesc;
               k := 0;
               FOR i := 0 TO 6 DO
                  IF k < 4
                  THEN BEGIN
                     s := [];
                     IF i > 3
                     THEN n := i-3
                     ELSE n := 0;
                     FOR j := n TO i-n DO
                        s := s + (l^.siteset[j] * r^.siteset[i-j]);
                     IF s <> []
                     THEN BEGIN
                        q^.siteset[k] := s;
                        k := k + 1;
                        counted := TRUE;
                        END
                     ELSE IF NOT counted
                        THEN numsteps[m] := numsteps[m] + weight[m];
                     END;
               FOR i := 2 DOWNTO 0 DO
                  FOR a := ala TO stop DO
                     IF a IN q^.siteset[i]
                     THEN FOR j := i+1 TO 3 DO
                        q^.siteset[j] := q^.siteset[j]
                                          + translate [a][j-i];
               s := [];
               FOR i := 0 TO 3 DO BEGIN
                  q^.siteset[i] := q^.siteset[i] - s;
                  s := s + q^.siteset[i];
                  END;
            END; (* fillin *)

         BEGIN (* postorder *)
            IF p <> NIL
            THEN BEGIN
               postorder (p^.leftdesc);
               postorder (p^.rtdesc);
               IF NOT p^.tip
               THEN fillin (p);
               END;
         END; (* postorder *)

      BEGIN (* evaluate *)
         FOR m := 1 TO chars DO BEGIN
            FOR i := 1 TO spp DO
               treenode[i]^.siteset := translate [treenode[i]^.seq[m]];
            numsteps[m] := 0;
            postorder (root);
            END;
         sum := 0;
         FOR i := 1 TO chars DO BEGIN
            steps := numsteps[i];
            IF steps <= threshold
            THEN sum := sum + steps
            ELSE sum := sum + threshold;
            END;
         like := -sum;
      END; (* evaluate *)

      PROCEDURE addpreorder (p, item : pointer);
         (* traverses a binary tree, calling PROCEDURE tryadd
            at a node before calling tryadd at its descendants *)

         PROCEDURE tryadd (p: pointer);
            (* temporarily adds one fork and one tip to the tree.
               if the location where they are added yields greater
               "likelihood" than other locations tested up to that
               time, then keeps that location as there *)
            VAR nufork : pointer;
         BEGIN (* tryadd *)
            nufork := treenode[spp+i-1];
            add (p, item, nufork);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            evaluate (root);
            remove (item, nufork);
            IF nufork = root THEN root := p;
            IF like > bestyet THEN BEGIN
               bestyet := like;
               there := p;
               besteps := numsteps;
               END;
         END; (* tryadd *)

      BEGIN (* addpreorder *)
         IF p <> NIL
         THEN BEGIN
            tryadd (p);
            addpreorder (p^.leftdesc, item);
            addpreorder (p^.rtdesc, item);
            END;
      END; (* addpreorder *)

      PROCEDURE rearrange (VAR r: pointer);
         (* traverses the tree (preorder), finding any local
            rearrangement which decreases the number of steps.
            if traversal succeeds in increasing the tree's
            "likelihood", procedure rearrange runs traversal again *)
         VAR success : BOOLEAN;

         PROCEDURE tryrearr (p : pointer);
            (* evaluates one rearrangement of the tree.
               if the new tree has greater "likelihood" than the old
               one sets success := TRUE and keeps the new tree.
               otherwise, restores the old tree *)
            VAR frombelow, whereto, forknode : pointer;
               oldlike : REAL;
         BEGIN (* tryrearr *)
            IF p^.ancestor <> NIL
            THEN BEGIN
               IF p^.ancestor^.ancestor <> NIL
               THEN BEGIN
                  oldlike := like;
                  IF p^.ancestor^.leftdesc = p
                  THEN frombelow := p^.ancestor^.rtdesc
                  ELSE frombelow := p^.ancestor^.leftdesc;
                  whereto := p^.ancestor^.ancestor;
                  forknode := p^.ancestor;
                  remove (p, forknode);
                  add (whereto, p, forknode);
                  IF r^.ancestor <> NIL THEN r := r^.ancestor;
                  evaluate (r);
                  IF like <= oldlike
                  THEN BEGIN
                     remove (p, forknode);
                     add (frombelow, p, forknode);
                     IF r^.ancestor <> NIL THEN r := r^.ancestor;
                     like := oldlike;
                     END
                  ELSE BEGIN
                     success := TRUE;
                     besteps := numsteps;
                     END;
                  END;
               END;
         END; (* tryrearr *)

         PROCEDURE repreorder (p : pointer);
            (* traverses a binary tree, calling procedure tryrearr
               at a node before calling tryrearr at its descendants *)
         BEGIN (* repreorder *)
            IF p <> NIL
            THEN BEGIN
               tryrearr (p);
               repreorder (p^.leftdesc);
               repreorder (p^.rtdesc);
               END;
         END; (* repreorder *)

      BEGIN (* rearrange *)
         success := TRUE;
         WHILE success DO BEGIN
            success := FALSE;
            repreorder (r);
            END;
      END; (* rearrange *)

      PROCEDURE treeread;
         (* read in user-defined tree and set it up *)
         VAR ch : CHAR;
            nextnode : INTEGER;

         PROCEDURE findch (c : CHAR);
            (* scan forward until find character c *)
         BEGIN (* findch *)
            IF ch <> c
            THEN REPEAT
                  READ(infile, ch);
               UNTIL ch = c;
         END; (* findch *)

         PROCEDURE addelement (VAR p : pointer);
            (* recursive procedure adds nodes to user-defined
               tree *)
            VAR q : pointer;
               n : INTEGER;
         BEGIN (* addelement *)
            REPEAT
               READ(infile, ch);
            UNTIL ch <> ' ';
            IF ch = '('
            THEN BEGIN
               nextnode := nextnode + 1;
               q := treenode[nextnode];
               addelement (q^.leftdesc);
               q^.leftdesc^.ancestor := q;
               findch (',');
               IF EOLN(infile) THEN READLN(infile);
               addelement (q^.rtdesc);
               q^.rtdesc^.ancestor := q;
               findch (')');
               p := q;
               IF NOT EOLN(infile) THEN READ(infile, ch);
               END
            ELSE BEGIN
               n := 0;
               REPEAT
                  n := 10*n + ORD(ch) - ORD('0');
                  READ(infile, ch);
               UNTIL NOT ((ch >= '0') AND (ch <= '9'));
               p := treenode[n];
               END;
         END; (* addelement *)

      BEGIN (* treeread *)
         root := treenode[spp + 1];
         nextnode := spp;
         root^.ancestor := NIL;
         addelement (root);
         READLN(infile);
      END; (* treeread *)

      PROCEDURE printree;
         (* prints out diagram of the tree *)
         CONST down = 2;
            over = 3;
         VAR tipx, tipy, i, rover : INTEGER;

         PROCEDURE coordinates (p : pointer);
            (* establishes coordinates of nodes *)
         BEGIN (* coordinates *)
            IF p^.tip
            THEN BEGIN
               p^.xcoord := tipx;
               p^.ycoord := tipy;
               tipx := tipx + rover;
               tipy := tipy + down;
               END
            ELSE BEGIN
               coordinates (p^.leftdesc);
               p^.xcoord := p^.leftdesc^.xcoord;
               coordinates (p^.rtdesc);
               p^.ycoord := p^.rtdesc^.ycoord;
               END;
         END; (* coordinates *)

         PROCEDURE drawline (i : INTEGER);
            (* draws one row of the tree diagram by moving up tree *)
            VAR p, q : pointer;
               n, j : INTEGER;
         BEGIN (* drawline *)
            p :=root;
            WRITE(OUTPUT, '   ');
            REPEAT
               IF p^.leftdesc^.ycoord >= i
               THEN p := p^.leftdesc
               ELSE BEGIN
                  q := p^.rtdesc;
                  n := q^.xcoord - p^.xcoord;
                  IF p^.ycoord = i
                  THEN BEGIN
                     IF (p^.ycoord = (tipy - down))
                        AND (p^.xcoord = 0)
                     THEN WRITE(OUTPUT, 'L')
                     ELSE WRITE(OUTPUT, '*');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, '-');
                     END
                  ELSE BEGIN
                     WRITE(OUTPUT, '!');
                     FOR j := 1 TO n-1 DO WRITE(OUTPUT, ' ');
                     END;
                  p := q;
                  END;
            UNTIL p^.tip;
            IF p^.ycoord = i
            THEN FOR j := 1 TO nmlngth DO
                       WRITE(OUTPUT, name[p^.index, j]);
            WRITELN(OUTPUT);
         END; (* drawline *)

      BEGIN (* printree *)
         tipx := 0;
         tipy := 1;
         rover := 100 DIV spp;
         IF rover > over THEN rover := over;
         coordinates (root);
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO tipy-down DO drawline (i);
         WRITELN(OUTPUT, ' ');
         WRITE(OUTPUT, '   remember: ');
         IF didreroot
         THEN WRITE(OUTPUT, '(although rooted by outgroup) ');
         WRITELN(OUTPUT, 'this is an unrooted tree!');
         WRITELN(OUTPUT);
      END; (* printree *)

      PROCEDURE describe;
         (* prints number of steps and table of numbers of steps in
            each site *)
         VAR i, j, k : INTEGER;
      BEGIN (* describe *)
         WRITELN(OUTPUT);
         WRITELN(OUTPUT, ' requires a total of ',-like:10:3);
         WRITELN(OUTPUT);
         IF weights THEN WRITE(OUTPUT, ' weighted');
         WRITELN(OUTPUT, ' steps in each position:');
         WRITE(OUTPUT, '       ');
         FOR i := 0 TO 9 DO WRITE(OUTPUT, i:4);
         WRITELN(OUTPUT);
         WRITE(OUTPUT, '      *------------------------------------');
         WRITELN(OUTPUT, '-----');
         FOR i := 0 TO (chars DIV 10) DO BEGIN
            WRITE(OUTPUT, 10*i:6);
            WRITE(OUTPUT, '!');
            FOR j := 0 TO 9 DO BEGIN
               k := 10*i + j;
               IF (k = 0) OR (k > chars)
               THEN WRITE(OUTPUT, '    ')
               ELSE WRITE(OUTPUT, besteps[k]:4);
               END;
            WRITELN(OUTPUT);
            END;
         WRITELN(OUTPUT);
         WRITELN(OUTPUT);
      END; (* describe *)

      PROCEDURE reroot (outgroup : pointer);
         (* reorients tree, putting outgroup in desired position.
            Written by Kent Fiala *)
         VAR vertex, oldvertex, ante, otherbranch : pointer;
            left : BOOLEAN;
      BEGIN (* reroot *)
         IF NOT (outgroup^.ancestor = root)
         THEN BEGIN
            oldvertex := outgroup;
            vertex := outgroup^.ancestor;
            ante := vertex^.ancestor;
            IF vertex^.rtdesc = outgroup
            THEN BEGIN
               vertex^.rtdesc := vertex^.leftdesc;
               vertex^.leftdesc := outgroup;
               END;
            REPEAT
               IF oldvertex = vertex^.leftdesc
               THEN BEGIN
                  vertex^.leftdesc := ante;
                  left := TRUE;
                  END
               ELSE BEGIN
                  vertex^.rtdesc := ante;
                  left := FALSE;
                  END;
               vertex^.ancestor := oldvertex;
               oldvertex := vertex;
               vertex := ante;
               ante := vertex^.ancestor
            UNTIL vertex = root;
            IF root^.leftdesc = oldvertex
            THEN otherbranch := root^.rtdesc
            ELSE otherbranch := root^.leftdesc;
            IF left
            THEN oldvertex^.leftdesc := otherbranch
            ELSE oldvertex^.rtdesc := otherbranch;
            otherbranch^.ancestor := oldvertex;
            root^.rtdesc := outgroup^.ancestor;
            root^.leftdesc := outgroup;
            outgroup^.ancestor^.ancestor := root;
            outgroup^.ancestor := root;
            END;
      END; (* reroot *)

   BEGIN (* maketree *)
      setup;
      IF NOT usertree
      THEN BEGIN
         add (treenode[2], treenode[1], treenode[spp+1]);
         root := treenode[spp+1];
         FOR i:= 3 TO spp DO BEGIN
            bestyet := -32000;
            there := root;
            item := treenode[i];
            addpreorder (root, item);
            add (there, treenode[i], treenode[spp+i-1]);
            IF root^.ancestor <> NIL THEN root := root^.ancestor;
            like := bestyet;
            rearrange (root);
            IF (NOT brief) OR (i = spp)
            THEN BEGIN
               didreroot := outgropt AND (i = spp);
               IF didreroot
               THEN reroot (treenode[outgrno]);
               printree;
               describe;
               END;
            END;
         END
      ELSE BEGIN
         READLN(infile, numtrees);
         WRITELN(OUTPUT, ' User-defined trees:');
         WRITELN(OUTPUT, ' ');
         FOR i := 1 TO numtrees DO BEGIN
            treeread;
            evaluate (root);
            besteps := numsteps;
            printree;
            describe;
            END;
         END;
   END; (* maketree *)

BEGIN (* Protein parsimony by uphill search *)
   (* reads in spp, chars, and the data. Then calls maketree to
      construct the tree *)
   RESET(infile);
   doinput;
   maketree;
  (*  close(infile);  -- may need this here *)
END. (* Protein parsimony by uphill search *)
